// caller must define at compile time:
// n
// k
// c: choose c columns of H anew in each iteration
// m: number of tries for Z is m
// l: #Z_1 = l; #Z_2 = l; ... #Z_m = l; currently must have l<=24
// x: #X = x; #Y = k-x

// caller must pass in as run-time arg:
int w = 10;

// not currently adjustable:
// p=2: try all A subseteq X with #A = p
// q=2: try all B subseteq Y with #B = q

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

long long job = 0;
long long seed = 0;
long long maxiterations = -1;

typedef unsigned long long uint64;
typedef unsigned int uint32;

#ifdef __amd64__

#define limb uint64 __attribute__ ((aligned (16)))
#define LIMBBITS 64
#define MASK1 0x5555555555555555ULL
#define MASK2 0x3333333333333333ULL
#define MASK3 0x0f0f0f0f0f0f0f0fULL
#define MASK4 0x00ff00ff00ff00ffULL
#define MASK5 0x0000ffff0000ffffULL
#define MASK6 0x00000000ffffffffULL

long long cpucycles(void)
{
  unsigned long long result;
  asm volatile(".byte 15;.byte 49;shlq $32,%%rdx;orq %%rdx,%%rax"
    : "=a" (result) ::  "%rdx");
  return result;
}

#else

#define limb uint32 __attribute__ ((aligned (16)))
#define LIMBBITS 32
#define MASK1 0x55555555ULL
#define MASK2 0x33333333ULL
#define MASK3 0x0f0f0f0fULL
#define MASK4 0x00ff00ffULL
#define MASK5 0x0000ffffULL

long long cpucycles(void)
{
  long long result;
  asm volatile(".byte 15;.byte 49" : "=A" (result));
  return result;
}

#endif

#define GLIMBS ((n + LIMBBITS - 1)/LIMBBITS)
#define HLIMBS ((n - k + LIMBBITS - 1)/LIMBBITS)

limb G[k][GLIMBS]; // G[0],G[1],...,G[k-1] are basis for set of codewords
limb H[n][HLIMBS]; // H~[0],H~[1],...,H~[n-k-1] are basis for set of parity checks

#define SEENLIMBS ((1 << l) / LIMBBITS)
limb seenahash[SEENLIMBS];
limb seenbhash[SEENLIMBS];

#define SET(v,p) ((v)[(p) / LIMBBITS] |= (((limb) 1) << ((p) & (LIMBBITS - 1))))
#define CLEAR(v,p) ((v)[(p) / LIMBBITS] &= ~(((limb) 1) << ((p) & (LIMBBITS - 1))))
#define ISSET(v,p) ((v)[(p) / LIMBBITS] & (((limb) 1) << ((p) & (LIMBBITS - 1))))

int pivot[k];
int nonpivot[n - k];
// originally: G[i] is only basis element where bit pivot[i] is set.
// later: nonpivot[i] are the selected columns;
// H~[i] is only basis element where bit nonpivot[i] is set.

int ispivot[n];
// ispivot[pivot[i]] = 1; ispivot[nonpivot[i]] = 0

int myrandom(int mod)
{
  int result = (random() + (random() >> 16)) % mod;
  while (result < 0) result += mod;
  while (result >= mod) result -= mod;
  return result;
}

void G_init(void)
{
  int i;
  int j;
  int r;
  char ch;

  for (i = 0;i < k;++i)
    for (j = 0;j < GLIMBS;++j)
      G[i][j] = 0;

  for (i = 0;i < k;++i) {
    if (i == 0) {
      if (scanf(" %c",&ch) != 1) exit(111); if (ch != '[') exit(100);
    }
    if (scanf(" %c",&ch) != 1) exit(111); if (ch != '[') exit(100);
    for (j = 0;j < n;++j) {
      if (scanf("%d",&r) != 1) exit(111);
      if (r) SET(G[i],j);
    }
    if (scanf(" %c",&ch) != 1) exit(111); if (ch != ']') exit(100);
    if (i == k - 2) {
      if (scanf(" %c",&ch) != 1) exit(111); if (ch != ']') exit(100);
    }
  }
}

void GH_reduceinit(void)
{
  int i;
  int j;
  int r;
  int h;

  for (i = 0;i < n;++i)
    ispivot[i] = 0;
  for (r = 0;r < k;++r) {
    for (pivot[r] = 0;pivot[r] < n;++pivot[r]) {
      if (ispivot[pivot[r]]) continue;
      if (!ISSET(G[r],pivot[r])) continue;
      ispivot[pivot[r]] = 1;
      for (i = 0;i < k;++i) {
        if (i == r) continue;
        if (!ISSET(G[i],pivot[r])) continue;
        for (j = 0;j < GLIMBS;++j) G[i][j] ^= G[r][j];
      }
      break;
    }
    if (pivot[r] == n) { printf("not full rank\n"); exit(100); }
  }
  h = 0;
  for (i = 0;i < n;++i)
    if (!ispivot[i]) {
      nonpivot[h] = i;
      SET(H[i],h);
      for (j = 0;j < k;++j)
        if (ISSET(G[j],i))
          SET(H[pivot[j]],h);
      ++h;
    }
}

void newcols(void)
{
  int i;
  int j;
  int r;
  int s;
  int t;
  int u;
  int replacing[c];

  i = 0;
  j = 0;
  for (i = 0;i < c;++i) {
    while (myrandom(n - k - j) >= c - i) ++j;
    replacing[i] = j++;
  }
  for (i = 0;i < c;++i) {
    do
      j = myrandom(k);
    while (!ISSET(H[pivot[j]],replacing[i]));
    r = pivot[j];
    s = nonpivot[replacing[i]];
    for (t = 0;t < n;++t) {
      if (!ISSET(H[t],replacing[i])) continue;
      if (t == r) continue;
      for (u = 0;u < HLIMBS;++u) {
        if (u == replacing[i] / LIMBBITS) continue;
        H[t][u] ^= H[r][u];
      }
      u = replacing[i] / LIMBBITS;
      H[t][u] ^= H[r][u] & ~(((limb) 1) << (replacing[i] & (LIMBBITS - 1)));
    }
    for (u = 0;u < n - k;++u) {
      if (u == replacing[i]) continue;
      CLEAR(H[r],u);
    }
    nonpivot[replacing[i]] = r;
    pivot[j] = s;
    ispivot[r] = 0;
    ispivot[s] = 1;
  }
}

long long cyclesnewcols = 0;
long long cyclessearch = 0;
long long cyclestotal = 0;
long long cyclesextract = 0;
long long cyclessort = 0;
long long cyclescoll = 0;
long long cyclespairs = 0;

long long numiterations = 0;

long long aij[x][x];
long long bij[k - x][k - x];

void search_init(void)
{
  int i;
  int j;

  for (i = 0;i < x - 1;++i)
    for (j = i + 1;j < x;++j) {
      aij[i][j] = i;
      aij[i][j] <<= 16;
      aij[i][j] |= j;
      aij[i][j] <<= 32;
    }
  for (i = 0;i < k - x - 1;++i)
    for (j = i + 1;j < k - x;++j) {
      bij[i][j] = i;
      bij[i][j] <<= 16;
      bij[i][j] |= j;
      bij[i][j] <<= 32;
    }
}

static uint32 a1hash[x];
static uint32 a1hashdiff[x];
static uint32 b1hash[k - x];
static uint32 b1hashdiff[k - x];
static uint64 ahash[x * (x - 1) / 2];
static uint64 sortahash[x * (x - 1) / 2];
static uint64 bhash[(k - x) * (k - x - 1) / 2];
static uint64 sortbhash[(k - x) * (k - x - 1) / 2];
static int amax;
static int bmax;

void dopairs(void)
{
  int i;
  int j;
  uint32 hash;
  int a;
  int b;

  for (i = 0;i < SEENLIMBS;++i) seenbhash[i] = 0;
  for (i = 0;i < SEENLIMBS;++i) seenahash[i] = 0;

  for (i = 0;i < x - 1;++i) {
    j = i + 1;
    hash = a1hashdiff[i];
    if (1 & (x - j)) {
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
    }
    if (2 & (x - j)) {
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
    }
    while (j < x) {
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
      SET(seenahash,hash);
      hash ^= a1hashdiff[j++];
    }
  }

  b = 0;
  for (i = 0;i < k - x - 1;++i) {
    j = i + 1;
    hash = b1hashdiff[i];
    if (1 & (k - x - j)) {
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
    }
    if (2 & (k - x - j)) {
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
    }
    while (j < k - x) {
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
      if (ISSET(seenahash,hash)) { SET(seenbhash,hash); bhash[b++] = bij[i][j] ^ hash; }
      hash ^= b1hashdiff[j++];
    }
  }
  bmax = b;

  a = 0;
  for (i = 0;i < x - 1;++i) {
    j = i + 1;
    hash = a1hashdiff[i];
    if (1 & (x - j)) {
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
    }
    if (2 & (x - j)) {
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
    }
    while (j < x) {
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
      if (ISSET(seenbhash,hash)) { ahash[a++] = aij[i][j] ^ hash; }
      hash ^= a1hashdiff[j++];
    }
  }
  amax = a;
}

void search(void)
{
  int h;
  int i;
  int j;
  int r;
  int z;
  static int Z[m][l];
  static int X[x];
  static int Y[k - x];
  uint32 hash;
  int a;
  int b;
  int bmin;
  int bnext;
  static limb tmp[HLIMBS];
#define COUNT 64
#define COUNTBITS 6
  static uint32 count[64];
  int pos;

  j = 0;
  for (i = 0;i < m * l;++i) {
    while (myrandom(n - k - j) >= m * l - i) ++j;
    Z[i / l][i % l] = j++;
  }
  h = 0;
  j = 0;
  for (i = 0;i < x;++i) {
    while (myrandom(k - j) >= x - i) Y[h++] = j++;
    X[i] = j++;
  }
  while (h < k - x) Y[h++] = j++;

  for (z = 0;z < m;++z) {
    cyclesextract -= cpucycles();

    for (i = 0;i < x;++i) {
      hash = 0;
      for (r = 0;r < l;++r)
        hash |= ((H[pivot[X[i]]][Z[z][r] / LIMBBITS] >> (Z[z][r] & (LIMBBITS - 1))) & 1) << r;
      a1hash[i] = hash;
    }
    for (i = 1;i < x;++i)
      a1hashdiff[i - 1] = a1hash[i] ^ a1hash[i - 1];
    for (i = 0;i < k - x;++i) {
      hash = 0;
      for (r = 0;r < l;++r)
        hash |= ((H[pivot[Y[i]]][Z[z][r] / LIMBBITS] >> (Z[z][r] & (LIMBBITS - 1))) & 1) << r;
      b1hash[i] = hash;
    }
    for (i = 1;i < k - x;++i)
      b1hashdiff[i - 1] = b1hash[i] ^ b1hash[i - 1];

    cyclesextract += cpucycles();

    cyclespairs -= cpucycles();
    dopairs();
    cyclespairs += cpucycles();

    cyclessort -= cpucycles();

    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (a = 0;a < amax;++a) ++count[ahash[a] & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (a > 0) { pos = ahash[--a] & (COUNT - 1); sortahash[--count[pos]] = ahash[a]; }
    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (a = 0;a < amax;++a) ++count[(sortahash[a] >> COUNTBITS) & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (a > 0) { pos = (sortahash[--a] >> COUNTBITS) & (COUNT - 1); ahash[--count[pos]] = sortahash[a]; }
    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (a = 0;a < amax;++a) ++count[(ahash[a] >> (COUNTBITS * 2)) & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (a > 0) { pos = (ahash[--a] >> (COUNTBITS * 2)) & (COUNT - 1); sortahash[--count[pos]] = ahash[a]; }
    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (a = 0;a < amax;++a) ++count[(sortahash[a] >> (COUNTBITS * 3)) & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (a > 0) { pos = (sortahash[--a] >> (COUNTBITS * 3)) & (COUNT - 1); ahash[--count[pos]] = sortahash[a]; }

    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (b = 0;b < bmax;++b) ++count[bhash[b] & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (b > 0) { pos = bhash[--b] & (COUNT - 1); sortbhash[--count[pos]] = bhash[b]; }
    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (b = 0;b < bmax;++b) ++count[(sortbhash[b] >> COUNTBITS) & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (b > 0) { pos = (sortbhash[--b] >> COUNTBITS) & (COUNT - 1); bhash[--count[pos]] = sortbhash[b]; }
    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (b = 0;b < bmax;++b) ++count[(bhash[b] >> (COUNTBITS * 2)) & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (b > 0) { pos = (bhash[--b] >> (COUNTBITS * 2)) & (COUNT - 1); sortbhash[--count[pos]] = bhash[b]; }
    for (pos = 0;pos < COUNT;++pos) count[pos] = 0;
    for (b = 0;b < bmax;++b) ++count[(sortbhash[b] >> (COUNTBITS * 3)) & (COUNT - 1)];
    for (pos = 1;pos < COUNT;++pos) count[pos] += count[pos - 1];
    while (b > 0) { pos = (sortbhash[--b] >> (COUNTBITS * 3)) & (COUNT - 1); bhash[--count[pos]] = sortbhash[b]; }

    cyclessort += cpucycles();
    cyclescoll -= cpucycles();

    h = -1;
    bmin = 0;
    bnext = 0;
    for (a = 0;a < amax;++a) {
      if ((0xffffffff & ahash[a]) != h) {
        h = (0xffffffff & ahash[a]);
        bmin = bnext;
        while ((bmin < bmax) && ((0xffffffff & bhash[bmin]) < h)) ++bmin;
        bnext = bmin;
        while ((bnext < bmax) && ((0xffffffff & bhash[bnext]) == h)) ++bnext;
      }
      for (b = bmin;b < bnext;++b) {
        int ai = ahash[a] >> 48;
        int aj = 65535 & (ahash[a] >> 32);
        int bi = bhash[b] >> 48;
        int bj = 65535 & (bhash[b] >> 32);
        int weight = w - 4;

        for (r = 0;r < HLIMBS;++r) {
          limb x0
            = H[pivot[X[ai]]][r]
            ^ H[pivot[X[aj]]][r]
            ^ H[pivot[Y[bi]]][r]
            ^ H[pivot[Y[bj]]][r];
          limb x1 = (x0 & MASK1) + ((x0 >> 1) & MASK1);
          limb x2 = (x1 & MASK2) + ((x1 >> 2) & MASK2);
          limb x3 = (x2 & MASK3) + ((x2 >> 4) & MASK3);
          limb x4 = (x3 & MASK4) + ((x3 >> 8) & MASK4);
          limb x5 = (x4 & MASK5) + ((x4 >> 16) & MASK5);
#ifdef MASK6
          x5 = (x5 & MASK5) + ((x5 >> 32) & MASK6);
#endif
          weight -= x5;
          if (weight < 0) break;
        }
        if (weight == 0) {
          int bits[n];
          for (i = 0;i < n;++i) bits[i] = 0;
	  ++numiterations;
          fprintf(stderr,"job %lld seed %lld found by %lld iter; %.0f cycles/iter: %.0f newcols, %.0f search: %.0f extract, %.0f pairs, %.0f sort, %.0f coll\n"
	    ,job
	    ,seed
	    ,numiterations
            ,(cpucycles() + cyclestotal) * 1.0 / numiterations
            ,(cyclesnewcols) * 1.0 / numiterations
            ,(cpucycles() + cyclessearch) * 1.0 / numiterations
            ,(cyclesextract) * 1.0 / numiterations
            ,(cyclespairs) * 1.0 / numiterations
            ,(cyclessort) * 1.0 / numiterations
            ,(cpucycles() + cyclescoll) * 1.0 / numiterations
          );
          printf("positions");
#define PRINTBIT(j) printf(" %d",j); bits[j] = 1;
          PRINTBIT(pivot[X[ai]])
          PRINTBIT(pivot[X[aj]])
          PRINTBIT(pivot[Y[bi]])
          PRINTBIT(pivot[Y[bj]])
          for (r = 0;r < HLIMBS;++r) {
            tmp[r]
              = H[pivot[X[ai]]][r]
              ^ H[pivot[X[aj]]][r]
              ^ H[pivot[Y[bi]]][r]
              ^ H[pivot[Y[bj]]][r];
          }
          for (i = 0;i < n - k;++i)
            if (ISSET(tmp,i)) {
              PRINTBIT(nonpivot[i])
            }
          printf("\n");
          printf("sorted");
          for (i = 0;i < n;++i)
            if (bits[i])
              printf(" %d",i);
          printf("\n");
          printf("vector ");
          for (i = 0;i < n;++i) printf("%d",bits[i]);
          printf("\n");
	  exit(0);
        }
      }
    }
    cyclescoll += cpucycles();
  }
}

volatile long long terminating = 0;

void sigterm()
{
  terminating = 1;
}

void attack(void)
{
  long long terminatenow;

  search_init();
  cyclestotal -= cpucycles();
  for (;;) {
    cyclesnewcols -= cpucycles();
    newcols();
    cyclesnewcols += cpucycles();
    cyclessearch -= cpucycles();
    search();
    cyclessearch += cpucycles();
    ++numiterations;
    if (maxiterations >= 0)
      if (numiterations >= maxiterations)
        terminating = 1;
    terminatenow = terminating;
    if (terminatenow || !(numiterations & (numiterations - 1))) {
      fprintf(stderr,"job %lld seed %lld %s %lld iter; %.0f cycles/iter: %.0f newcols, %.0f search: %.0f extract, %.0f pairs, %.0f sort, %.0f coll\n"
	,job
	,seed
	,terminatenow ? "terminating after" : "finished"
	,numiterations
        ,(cpucycles() + cyclestotal) * 1.0 / numiterations
        ,(cyclesnewcols) * 1.0 / numiterations
        ,(cyclessearch) * 1.0 / numiterations
        ,(cyclesextract) * 1.0 / numiterations
        ,(cyclespairs) * 1.0 / numiterations
        ,(cyclessort) * 1.0 / numiterations
        ,(cyclescoll) * 1.0 / numiterations
      );
      fflush(stderr);
      if (terminatenow) return;
    }
  }
}

int main(int argc,char **argv)
{
  if (argv[1]) {
    w = atoll(argv[1]);
    if (argv[2]) {
      job = atoll(argv[2]);
      if (argv[3]) {
        seed = atoll(argv[3]);
        if (argv[4]) {
          maxiterations = atoll(argv[4]);
        }
      }
    }
  }

  fprintf(stderr,"job %lld seed %lld starting n %d k %d x %d l %d m %d c %d w %d maxiterations %lld\n",job,seed,n,k,x,l,m,c,w,maxiterations);

  srandom(seed);
  G_init();
  GH_reduceinit();
  signal(SIGTERM,sigterm);
  attack();
  return 0;
}
